@@ -38,7 +38,6 @@ static int  pygobject_clear(PyGObject *self);
 static PyObject * pyg_type_get_bases(GType gtype);
 static inline int pygobject_clear(PyGObject *self);
 static PyObject * pygobject_weak_ref_new(GObject *obj, PyObject *callback, PyObject *user_data);
-static PyObject * pygbinding_weak_ref_new(GObject *obj);
 static inline PyGObjectData * pyg_object_peek_inst_data(GObject *obj);
 static void pygobject_inherit_slots(PyTypeObject *type, PyObject *bases,
 				    gboolean check_for_present);
@@ -242,14 +241,13 @@ build_parameter_list(GObjectClass *class)
 	g_free(name);
     }
 
-    g_type_class_unref(class);
-
     if (props)
         g_free(props);
     
     return props_list;
 }
 
+
 static PyObject*
 PyGProps_getattro(PyGProps *self, PyObject *attr)
 {
@@ -266,12 +264,6 @@ PyGProps_getattro(PyGProps *self, PyObject *attr)
     }
 
     class = g_type_class_ref(self->gtype);
-    
-    if (!strcmp(attr_name, "__members__")) {
-        ret = build_parameter_list(class);
-        g_type_class_unref(class);
-	return ret;
-    }
 
     /* g_object_class_find_property recurses through the class hierarchy,
      * so the resulting pspec tells us the owner_type that owns the property
@@ -441,6 +433,25 @@ pygobject_props_get_iter(PyGProps *self)
     return (PyObject *) iter;
 }
 
+static PyObject*
+pygobject_props_dir(PyGProps *self)
+{
+    PyObject *ret;
+    GObjectClass *class;
+
+    class = g_type_class_ref (self->gtype);
+    ret = build_parameter_list (class);
+    g_type_class_unref (class);
+
+    return ret;
+}
+
+static PyMethodDef pygobject_props_methods[] = {
+    { "__dir__", (PyCFunction)pygobject_props_dir, METH_NOARGS},
+    { NULL, NULL, 0}
+};
+
+
 static Py_ssize_t
 PyGProps_length(PyGProps *self)
 {
@@ -1637,7 +1648,7 @@ pygobject_bind_property(PyGObject *self, PyObject *args)
 		return NULL;
 	}
 
-	return pygbinding_weak_ref_new(G_OBJECT (binding));
+	return pygobject_new (G_OBJECT (binding));
 }
 
 static PyObject *
@@ -2341,54 +2352,6 @@ pygobject_weak_ref_call(PyGObjectWeakRef *self, PyObject *args, PyObject *kw)
     }
 }
 
-
-/* -------------- GBinding Weak Reference ----------------- */
-
-/**
- * BindingWeakRef
- *
- * The BindingWeakRef object is used to manage GBinding objects within python
- * created through GObject.bind_property. It is a sub-class PyGObjectWeakRef so
- * that we can maintain the same reference counting semantics between Python
- * and GObject Binding objects. This gives explicit direct control of the
- * binding lifetime by using the "unbind" method on the BindingWeakRef object
- * along with implicit management based on the lifetime of the source or
- * target objects.
- */
-
-PYGLIB_DEFINE_TYPE("gi._gobject.GBindingWeakRef", PyGBindingWeakRef_Type, PyGObjectWeakRef);
-
-static PyObject *
-pygbinding_weak_ref_new(GObject *obj)
-{
-	PyGObjectWeakRef *self;
-
-	self = PyObject_GC_New(PyGObjectWeakRef, &PyGBindingWeakRef_Type);
-	self->callback = NULL;
-	self->user_data = NULL;
-	self->obj = obj;
-	g_object_weak_ref(self->obj, (GWeakNotify) pygobject_weak_ref_notify, self);
-	return (PyObject *) self;
-}
-
-static PyObject *
-pygbinding_weak_ref_unbind(PyGObjectWeakRef *self, PyObject *args)
-{
-    if (!self->obj) {
-        PyErr_SetString(PyExc_ValueError, "weak binding ref already unreffed");
-        return NULL;
-    }
-    g_object_unref(self->obj);
-    Py_INCREF(Py_None);
-    return Py_None;
-}
-
-static PyMethodDef pygbinding_weak_ref_methods[] = {
-    { "unbind", (PyCFunction)pygbinding_weak_ref_unbind, METH_NOARGS},
-    { NULL, NULL, 0}
-};
-
-
 static gpointer
 pyobject_copy(gpointer boxed)
 {
@@ -2462,6 +2425,7 @@ pygobject_object_register_types(PyObject *d)
 	"Python attributes.";
     PyGProps_Type.tp_traverse = (traverseproc)pygobject_props_traverse;
     PyGProps_Type.tp_iter = (getiterfunc)pygobject_props_get_iter;
+    PyGProps_Type.tp_methods = pygobject_props_methods;
     if (PyType_Ready(&PyGProps_Type) < 0)
         return;
 
@@ -2494,12 +2458,4 @@ pygobject_object_register_types(PyObject *d)
     if (PyType_Ready(&PyGObjectWeakRef_Type) < 0)
         return;
     PyDict_SetItemString(d, "GObjectWeakRef", (PyObject *) &PyGObjectWeakRef_Type);
-
-    PyGBindingWeakRef_Type.tp_flags = Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_GC;
-    PyGBindingWeakRef_Type.tp_doc = "A GBinding weak reference";
-    PyGBindingWeakRef_Type.tp_methods = pygbinding_weak_ref_methods;
-    PyGBindingWeakRef_Type.tp_base = &PyGObjectWeakRef_Type;
-    if (PyType_Ready(&PyGBindingWeakRef_Type) < 0)
-        return;
-    PyDict_SetItemString(d, "GBindingWeakRef", (PyObject *) &PyGBindingWeakRef_Type);
 }
