@@ -247,7 +247,8 @@ _pygi_marshal_from_py_void (PyGIInvokeState   *state,
                             PyGICallableCache *callable_cache,
                             PyGIArgCache      *arg_cache,
                             PyObject          *py_arg,
-                            GIArgument        *arg)
+                            GIArgument        *arg,
+                            gpointer          *cleanup_data)
 {
     g_warn_if_fail (arg_cache->transfer == GI_TRANSFER_NOTHING);
 
@@ -255,23 +256,16 @@ _pygi_marshal_from_py_void (PyGIInvokeState   *state,
         arg->v_pointer = NULL;
     } else if (PYGLIB_CPointer_Check(py_arg)) {
         arg->v_pointer = PYGLIB_CPointer_GetPointer (py_arg, NULL);
+    } else if (PYGLIB_PyLong_Check(py_arg) || PyLong_Check(py_arg)) {
+        arg->v_pointer = PyLong_AsVoidPtr (py_arg);
     } else {
-        /* NOTE: This will change to only allow integers and the deprecation
-         * warning will become a runtime exception. Using the following:
-         * arg->v_pointer = PyLong_AsVoidPtr (py_arg);
-         * See: https://bugzilla.gnome.org/show_bug.cgi?id=688081
-         */
-
-        if (!PYGLIB_PyLong_Check(py_arg) && !PyLong_Check(py_arg)) {
-            if (PyErr_WarnEx(PyGIDeprecationWarning,
-                             "Pointer arguments will be restricted to integers, capsules, and None. "
-                             "See: https://bugzilla.gnome.org/show_bug.cgi?id=683599",
-                             1))
-                return FALSE;
-        }
-        arg->v_pointer = py_arg;
+        PyErr_SetString(PyExc_ValueError,
+                        "Pointer arguments are restricted to integers, capsules, and None. "
+                        "See: https://bugzilla.gnome.org/show_bug.cgi?id=683599");
+        return FALSE;
     }
 
+    *cleanup_data = arg->v_pointer;
     return TRUE;
 }
 
@@ -416,7 +410,8 @@ _pygi_marshal_from_py_gtype (PyObject          *py_arg,
 
 static gboolean
 _pygi_marshal_from_py_utf8 (PyObject          *py_arg,
-                            GIArgument        *arg)
+                            GIArgument        *arg,
+                            gpointer          *cleanup_data)
 {
     gchar *string_;
 
@@ -445,12 +440,14 @@ _pygi_marshal_from_py_utf8 (PyObject          *py_arg,
     }
 
     arg->v_string = string_;
+    *cleanup_data = arg->v_string;
     return TRUE;
 }
 
 static gboolean
 _pygi_marshal_from_py_filename (PyObject          *py_arg,
-                                GIArgument        *arg)
+                                GIArgument        *arg,
+                                gpointer          *cleanup_data)
 {
     gchar *string_;
     GError *error = NULL;
@@ -484,6 +481,7 @@ _pygi_marshal_from_py_filename (PyObject          *py_arg,
         return FALSE;
     }
 
+    *cleanup_data = arg->v_string;
     return TRUE;
 }
 
@@ -634,7 +632,8 @@ gboolean
 _pygi_marshal_from_py_basic_type (PyObject   *object,   /* in */
                                   GIArgument *arg,      /* out */
                                   GITypeTag   type_tag,
-                                  GITransfer  transfer)
+                                  GITransfer  transfer,
+                                  gpointer   *cleanup_data /* out */)
 {
     switch (type_tag) {
         case GI_TYPE_TAG_VOID:
@@ -647,6 +646,7 @@ _pygi_marshal_from_py_basic_type (PyObject   *object,   /* in */
                     "See: https://bugzilla.gnome.org/show_bug.cgi?id=683599");
             } else {
                 arg->v_pointer = PyLong_AsVoidPtr (object);
+                *cleanup_data = arg->v_pointer;
             }
             break;
         case GI_TYPE_TAG_INT8:
@@ -690,10 +690,10 @@ _pygi_marshal_from_py_basic_type (PyObject   *object,   /* in */
             return _pygi_marshal_from_py_unichar (object, arg);
 
         case GI_TYPE_TAG_UTF8:
-            return _pygi_marshal_from_py_utf8 (object, arg);
+            return _pygi_marshal_from_py_utf8 (object, arg, cleanup_data);
 
         case GI_TYPE_TAG_FILENAME:
-            return _pygi_marshal_from_py_filename (object, arg);
+            return _pygi_marshal_from_py_filename (object, arg, cleanup_data);
 
         default:
             return FALSE;
@@ -710,12 +710,14 @@ _pygi_marshal_from_py_basic_type_cache_adapter (PyGIInvokeState   *state,
                                                 PyGICallableCache *callable_cache,
                                                 PyGIArgCache      *arg_cache,
                                                 PyObject          *py_arg,
-                                                GIArgument        *arg)
+                                                GIArgument        *arg,
+                                                gpointer          *cleanup_data)
 {
     return _pygi_marshal_from_py_basic_type (py_arg,
                                              arg,
                                              arg_cache->type_tag,
-                                             arg_cache->transfer);
+                                             arg_cache->transfer,
+                                             cleanup_data);
 }
 
 gboolean
@@ -723,10 +725,12 @@ _pygi_marshal_from_py_array (PyGIInvokeState   *state,
                              PyGICallableCache *callable_cache,
                              PyGIArgCache      *arg_cache,
                              PyObject          *py_arg,
-                             GIArgument        *arg)
+                             GIArgument        *arg,
+                             gpointer          *cleanup_data)
 {
     PyGIMarshalFromPyFunc from_py_marshaller;
     int i = 0;
+    int success_count = 0;
     Py_ssize_t length;
     gssize item_size;
     gboolean is_ptr_array;
@@ -760,10 +764,10 @@ _pygi_marshal_from_py_array (PyGIInvokeState   *state,
     item_size = sequence_cache->item_size;
     is_ptr_array = (sequence_cache->array_type == GI_ARRAY_TYPE_PTR_ARRAY);
     if (is_ptr_array) {
-        array_ = (GArray *)g_ptr_array_new ();
+        array_ = (GArray *)g_ptr_array_sized_new (length);
     } else {
         array_ = g_array_sized_new (sequence_cache->is_zero_terminated,
-                                    FALSE,
+                                    TRUE,
                                     item_size,
                                     length);
     }
@@ -786,8 +790,9 @@ _pygi_marshal_from_py_array (PyGIInvokeState   *state,
     }
 
     from_py_marshaller = sequence_cache->item_cache->from_py_marshaller;
-    for (i = 0; i < length; i++) {
-        GIArgument item;
+    for (i = 0, success_count = 0; i < length; i++) {
+        GIArgument item = {0};
+        gpointer item_cleanup_data = NULL;
         PyObject *py_item = PySequence_GetItem (py_arg, i);
         if (py_item == NULL)
             goto err;
@@ -796,8 +801,26 @@ _pygi_marshal_from_py_array (PyGIInvokeState   *state,
                                   callable_cache,
                                   sequence_cache->item_cache,
                                   py_item,
-                                 &item))
+                                 &item,
+                                 &item_cleanup_data)) {
+            Py_DECREF (py_item);
             goto err;
+        }
+        Py_DECREF (py_item);
+
+        if (item_cleanup_data != NULL && item_cleanup_data != item.v_pointer) {
+            /* We only support one level of data discrepancy between an items
+             * data and its cleanup data. This is because we only track a single
+             * extra cleanup data pointer per-argument and cannot track the entire
+             * array of items differing data and cleanup_data.
+             * For example, this would fail if trying to marshal an array of
+             * callback closures marked with SCOPE call type where the cleanup data
+             * is different from the items v_pointer, likewise an array of arrays.
+             */
+            PyErr_SetString(PyExc_RuntimeError, "Cannot cleanup item data for array due to "
+                                                "the items data its cleanup data being different.");
+            goto err;
+        }
 
         /* FIXME: it is much more efficent to have seperate marshaller
          *        for ptr arrays than doing the evaluation
@@ -805,7 +828,12 @@ _pygi_marshal_from_py_array (PyGIInvokeState   *state,
          */
         if (is_ptr_array) {
             g_ptr_array_add((GPtrArray *)array_, item.v_pointer);
+        } else if (sequence_cache->item_cache->is_pointer) {
+            /* if the item is a pointer, simply copy the pointer */
+            g_assert (item_size == sizeof (item.v_pointer));
+            g_array_insert_val (array_, i, item);
         } else if (sequence_cache->item_cache->type_tag == GI_TYPE_TAG_INTERFACE) {
+            /* Special case handling of flat arrays of gvalue/boxed/struct */
             PyGIInterfaceCache *item_iface_cache = (PyGIInterfaceCache *) sequence_cache->item_cache;
             GIBaseInfo *base_info = (GIBaseInfo *) item_iface_cache->interface_info;
             GIInfoType info_type = g_base_info_get_type (base_info);
@@ -816,58 +844,41 @@ _pygi_marshal_from_py_array (PyGIInvokeState   *state,
                 {
                     PyGIArgCache *item_arg_cache = (PyGIArgCache *)item_iface_cache;
                     PyGIMarshalCleanupFunc from_py_cleanup = item_arg_cache->from_py_cleanup;
-                    gboolean is_boxed = g_type_is_a (item_iface_cache->g_type, G_TYPE_BOXED);
-                    gboolean is_gvalue = item_iface_cache->g_type == G_TYPE_VALUE;
-                    gboolean is_gvariant = item_iface_cache->g_type == G_TYPE_VARIANT;
-                    
-                    if (is_gvariant) {
-                        /* Item size will always be that of a pointer,
-                         * since GVariants are opaque hence always passed by ref */
-                        g_assert (item_size == sizeof (item.v_pointer));
-                        g_array_insert_val (array_, i, item.v_pointer);
-                    } else if (is_gvalue) {
+
+                    if (g_type_is_a (item_iface_cache->g_type, G_TYPE_VALUE)) {
+                        /* Special case GValue flat arrays to properly init and copy the contents. */
                         GValue* dest = (GValue*) (array_->data + (i * item_size));
-                        memset (dest, 0, item_size);
                         if (item.v_pointer != NULL) {
+                            memset (dest, 0, item_size);
                             g_value_init (dest, G_VALUE_TYPE ((GValue*) item.v_pointer));
                             g_value_copy ((GValue*) item.v_pointer, dest);
                         }
-                        /* we free the original copy already, the new one is a plain struct
-                         * in an array. _pygi_marshal_cleanup_from_py_array() does not free it again */
-                        if (from_py_cleanup)
-                            from_py_cleanup (state, item_arg_cache, item.v_pointer, TRUE);
-                    } else if (!is_boxed) {
-                        /* HACK: Gdk.Atom is merely an integer wrapped in a pointer,
-                         * so we must not dereference it; just copy the pointer
-                         * value, and don't attempt to free it. TODO: find out
-                         * if there are other data types with similar behaviour
-                         * and generalize. */
-                        if (g_strcmp0 (item_iface_cache->type_name, "Gdk.Atom") == 0) {
-                            g_assert (item_size == sizeof (item.v_pointer));
-                            memcpy (array_->data + (i * item_size), &item.v_pointer, item_size);
-                        } else {
-                            memcpy (array_->data + (i * item_size), item.v_pointer, item_size);
+                        /* Manually increment the length because we are manually setting the memory. */
+                        array_->len++;
 
-                            if (from_py_cleanup)
-                                from_py_cleanup (state, item_arg_cache, item.v_pointer, TRUE);
-                        }
-                    } else if (is_boxed && !item_iface_cache->arg_cache.is_pointer) {
-                        /* The array elements are not expected to be pointers, but the
-                         * elements obtained are boxed pointers themselves, so insert
-                         * the pointed to data.
-                         */
-                        g_array_insert_vals (array_, i, item.v_pointer, 1);
                     } else {
-                        g_array_insert_val (array_, i, item);
+                        /* Handles flat arrays of boxed or struct types. */
+                        g_array_insert_vals (array_, i, item.v_pointer, 1);
                     }
+
+                    /* Cleanup any memory left by the per-item marshaler because
+                     * _pygi_marshal_cleanup_from_py_array will not know about this
+                     * due to "item" being a temporarily marshaled value done on the stack.
+                     */
+                    if (from_py_cleanup)
+                        from_py_cleanup (state, item_arg_cache, py_item, item_cleanup_data, TRUE);
+
                     break;
                 }
                 default:
                     g_array_insert_val (array_, i, item);
             }
         } else {
+            /* default value copy of a simple type */
             g_array_insert_val (array_, i, item);
         }
+
+        success_count++;
         continue;
 err:
         if (sequence_cache->item_cache->from_py_cleanup != NULL) {
@@ -875,11 +886,19 @@ err:
             PyGIMarshalCleanupFunc cleanup_func =
                 sequence_cache->item_cache->from_py_cleanup;
 
-            for(j = 0; j < i; j++) {
-                cleanup_func (state,
-                              sequence_cache->item_cache,
-                              g_array_index (array_, gpointer, j),
-                              TRUE);
+            /* Only attempt per item cleanup on pointer items */
+            if (sequence_cache->item_cache->is_pointer) {
+                for(j = 0; j < success_count; j++) {
+                    PyObject *py_item = PySequence_GetItem (py_arg, j);
+                    cleanup_func (state,
+                                  sequence_cache->item_cache,
+                                  py_item,
+                                  is_ptr_array ?
+                                          g_ptr_array_index ((GPtrArray *)array_, j) :
+                                          g_array_index (array_, gpointer, j),
+                                  TRUE);
+                    Py_DECREF (py_item);
+                }
             }
         }
 
@@ -895,7 +914,7 @@ array_success:
     if (sequence_cache->len_arg_index >= 0) {
         /* we have an child arg to handle */
         PyGIArgCache *child_cache =
-            callable_cache->args_cache[sequence_cache->len_arg_index];
+            _pygi_callable_cache_get_arg (callable_cache, sequence_cache->len_arg_index);
 
         if (child_cache->direction == PYGI_DIRECTION_BIDIRECTIONAL) {
             gint *len_arg = (gint *)state->in_args[child_cache->c_arg_index].v_pointer;
@@ -919,14 +938,34 @@ array_success:
     }
 
     if (sequence_cache->array_type == GI_ARRAY_TYPE_C) {
+        /* In the case of GI_ARRAY_C, we give the data directly as the argument
+         * but keep the array_ wrapper as cleanup data so we don't have to find
+         * it's length again.
+         */
         arg->v_pointer = array_->data;
-        g_array_free (array_, FALSE);
-        /* remember the originally allocated array in args_data, as args and
-         * in_args get changed for (inout) arguments */
-        if (arg_cache->transfer == GI_TRANSFER_NOTHING)
-            state->args_data[arg_cache->c_arg_index] = arg->v_pointer;
+
+        if (arg_cache->transfer == GI_TRANSFER_EVERYTHING) {
+            g_array_free (array_, FALSE);
+            *cleanup_data = NULL;
+        } else {
+            *cleanup_data = array_;
+        }
     } else {
         arg->v_pointer = array_;
+
+        if (arg_cache->transfer == GI_TRANSFER_NOTHING) {
+            /* Free everything in cleanup. */
+            *cleanup_data = array_;
+        } else if (arg_cache->transfer == GI_TRANSFER_CONTAINER) {
+            /* Make a shallow copy so we can free the elements later in cleanup
+             * because it is possible invoke will free the list before our cleanup. */
+            *cleanup_data = is_ptr_array ?
+                    (gpointer)g_ptr_array_ref ((GPtrArray *)array_) :
+                    (gpointer)g_array_ref (array_);
+        } else { /* GI_TRANSFER_EVERYTHING */
+            /* No cleanup, everything is given to the callee. */
+            *cleanup_data = NULL;
+        }
     }
 
     return TRUE;
@@ -937,7 +976,8 @@ _pygi_marshal_from_py_glist (PyGIInvokeState   *state,
                              PyGICallableCache *callable_cache,
                              PyGIArgCache      *arg_cache,
                              PyObject          *py_arg,
-                             GIArgument        *arg)
+                             GIArgument        *arg,
+                             gpointer          *cleanup_data)
 {
     PyGIMarshalFromPyFunc from_py_marshaller;
     int i;
@@ -971,7 +1011,8 @@ _pygi_marshal_from_py_glist (PyGIInvokeState   *state,
 
     from_py_marshaller = sequence_cache->item_cache->from_py_marshaller;
     for (i = 0; i < length; i++) {
-        GIArgument item;
+        GIArgument item = {0};
+        gpointer item_cleanup_data = NULL;
         PyObject *py_item = PySequence_GetItem (py_arg, i);
         if (py_item == NULL)
             goto err;
@@ -980,9 +1021,11 @@ _pygi_marshal_from_py_glist (PyGIInvokeState   *state,
                                   callable_cache,
                                   sequence_cache->item_cache,
                                   py_item,
-                                 &item))
+                                 &item,
+                                 &item_cleanup_data))
             goto err;
 
+        Py_DECREF (py_item);
         list_ = g_list_prepend (list_, _pygi_arg_to_hash_pointer (&item, sequence_cache->item_cache->type_tag));
         continue;
 err:
@@ -991,12 +1034,25 @@ err:
             PyGIMarshalCleanupFunc cleanup = sequence_cache->item_cache->from_py_cleanup;
         }
         */
+        Py_DECREF (py_item);
         g_list_free (list_);
         _PyGI_ERROR_PREFIX ("Item %i: ", i);
         return FALSE;
     }
 
     arg->v_pointer = g_list_reverse (list_);
+
+    if (arg_cache->transfer == GI_TRANSFER_NOTHING) {
+        /* Free everything in cleanup. */
+        *cleanup_data = arg->v_pointer;
+    } else if (arg_cache->transfer == GI_TRANSFER_CONTAINER) {
+        /* Make a shallow copy so we can free the elements later in cleanup
+         * because it is possible invoke will free the list before our cleanup. */
+        *cleanup_data = g_list_copy (arg->v_pointer);
+    } else { /* GI_TRANSFER_EVERYTHING */
+        /* No cleanup, everything is given to the callee. */
+        *cleanup_data = NULL;
+    }
     return TRUE;
 }
 
@@ -1005,7 +1061,8 @@ _pygi_marshal_from_py_gslist (PyGIInvokeState   *state,
                               PyGICallableCache *callable_cache,
                               PyGIArgCache      *arg_cache,
                               PyObject          *py_arg,
-                              GIArgument        *arg)
+                              GIArgument        *arg,
+                              gpointer          *cleanup_data)
 {
     PyGIMarshalFromPyFunc from_py_marshaller;
     int i;
@@ -1038,7 +1095,8 @@ _pygi_marshal_from_py_gslist (PyGIInvokeState   *state,
 
     from_py_marshaller = sequence_cache->item_cache->from_py_marshaller;
     for (i = 0; i < length; i++) {
-        GIArgument item;
+        GIArgument item = {0};
+        gpointer item_cleanup_data = NULL;
         PyObject *py_item = PySequence_GetItem (py_arg, i);
         if (py_item == NULL)
             goto err;
@@ -1047,9 +1105,11 @@ _pygi_marshal_from_py_gslist (PyGIInvokeState   *state,
                              callable_cache,
                              sequence_cache->item_cache,
                              py_item,
-                            &item))
+                            &item,
+                            &item_cleanup_data))
             goto err;
 
+        Py_DECREF (py_item);
         list_ = g_slist_prepend (list_, _pygi_arg_to_hash_pointer (&item, sequence_cache->item_cache->type_tag));
         continue;
 err:
@@ -1059,12 +1119,26 @@ err:
         }
         */
 
+        Py_DECREF (py_item);
         g_slist_free (list_);
         _PyGI_ERROR_PREFIX ("Item %i: ", i);
         return FALSE;
     }
 
     arg->v_pointer = g_slist_reverse (list_);
+
+    if (arg_cache->transfer == GI_TRANSFER_NOTHING) {
+        /* Free everything in cleanup. */
+        *cleanup_data = arg->v_pointer;
+    } else if (arg_cache->transfer == GI_TRANSFER_CONTAINER) {
+        /* Make a shallow copy so we can free the elements later in cleanup
+         * because it is possible invoke will free the list before our cleanup. */
+        *cleanup_data = g_slist_copy (arg->v_pointer);
+    } else { /* GI_TRANSFER_EVERYTHING */
+        /* No cleanup, everything is given to the callee. */
+        *cleanup_data = NULL;
+    }
+
     return TRUE;
 }
 
@@ -1073,7 +1147,8 @@ _pygi_marshal_from_py_ghash (PyGIInvokeState   *state,
                              PyGICallableCache *callable_cache,
                              PyGIArgCache      *arg_cache,
                              PyObject          *py_arg,
-                             GIArgument        *arg)
+                             GIArgument        *arg,
+                             gpointer          *cleanup_data)
 {
     PyGIMarshalFromPyFunc key_from_py_marshaller;
     PyGIMarshalFromPyFunc value_from_py_marshaller;
@@ -1136,6 +1211,8 @@ _pygi_marshal_from_py_ghash (PyGIInvokeState   *state,
 
     for (i = 0; i < length; i++) {
         GIArgument key, value;
+        gpointer key_cleanup_data = NULL;
+        gpointer value_cleanup_data = NULL;
         PyObject *py_key = PyList_GET_ITEM (py_keys, i);
         PyObject *py_value = PyList_GET_ITEM (py_values, i);
         if (py_key == NULL || py_value == NULL)
@@ -1145,14 +1222,16 @@ _pygi_marshal_from_py_ghash (PyGIInvokeState   *state,
                                       callable_cache,
                                       hash_cache->key_cache,
                                       py_key,
-                                     &key))
+                                     &key,
+                                     &key_cleanup_data))
             goto err;
 
         if (!value_from_py_marshaller ( state,
                                         callable_cache,
                                         hash_cache->value_cache,
                                         py_value,
-                                       &value))
+                                       &value,
+                                       &value_cleanup_data))
             goto err;
 
         g_hash_table_insert (hash_,
@@ -1171,6 +1250,21 @@ err:
     }
 
     arg->v_pointer = hash_;
+
+    if (arg_cache->transfer == GI_TRANSFER_NOTHING) {
+        /* Free everything in cleanup. */
+        *cleanup_data = arg->v_pointer;
+    } else if (arg_cache->transfer == GI_TRANSFER_CONTAINER) {
+        /* Make a shallow copy so we can free the elements later in cleanup
+         * because it is possible invoke will free the list before our cleanup. */
+        *cleanup_data = g_hash_table_ref (arg->v_pointer);
+    } else { /* GI_TRANSFER_EVERYTHING */
+        /* No cleanup, everything is given to the callee.
+         * Note that the keys and values will leak for transfer everything because
+         * we do not use g_hash_table_new_full and set key/value_destroy_func. */
+        *cleanup_data = NULL;
+    }
+
     return TRUE;
 }
 
@@ -1179,7 +1273,8 @@ _pygi_marshal_from_py_gerror (PyGIInvokeState   *state,
                               PyGICallableCache *callable_cache,
                               PyGIArgCache      *arg_cache,
                               PyObject          *py_arg,
-                              GIArgument        *arg)
+                              GIArgument        *arg,
+                              gpointer          *cleanup_data)
 {
     PyErr_Format (PyExc_NotImplementedError,
                   "Marshalling for GErrors is not implemented");
@@ -1245,7 +1340,8 @@ _pygi_marshal_from_py_interface_callback (PyGIInvokeState   *state,
                                           PyGICallableCache *callable_cache,
                                           PyGIArgCache      *arg_cache,
                                           PyObject          *py_arg,
-                                          GIArgument        *arg)
+                                          GIArgument        *arg,
+                                          gpointer          *cleanup_data)
 {
     GICallableInfo *callable_info;
     PyGICClosure *closure;
@@ -1257,12 +1353,17 @@ _pygi_marshal_from_py_interface_callback (PyGIInvokeState   *state,
     callback_cache = (PyGICallbackCache *)arg_cache;
 
     if (callback_cache->user_data_index > 0) {
-        user_data_cache = callable_cache->args_cache[callback_cache->user_data_index];
+        user_data_cache = _pygi_callable_cache_get_arg (callable_cache, callback_cache->user_data_index);
         if (user_data_cache->py_arg_index < state->n_py_in_args) {
             /* py_user_data is a borrowed reference. */
             py_user_data = PyTuple_GetItem (state->py_in_args, user_data_cache->py_arg_index);
             if (!py_user_data)
                 return FALSE;
+            /* NULL out user_data if it was not supplied and the default arg placeholder
+             * was used instead.
+             */
+            if (py_user_data == _PyGIDefaultArgPlaceholder)
+                py_user_data = NULL;
         }
     }
 
@@ -1305,7 +1406,7 @@ _pygi_marshal_from_py_interface_callback (PyGIInvokeState   *state,
      * later on in _pygi_destroy_notify_callback_closure.
      */
     if (callback_cache->destroy_notify_index > 0) {
-        destroy_cache = callable_cache->args_cache[callback_cache->destroy_notify_index];
+        destroy_cache = _pygi_callable_cache_get_arg (callable_cache, callback_cache->destroy_notify_index);
     }
 
     if (destroy_cache) {
@@ -1327,10 +1428,8 @@ _pygi_marshal_from_py_interface_callback (PyGIInvokeState   *state,
         }
     }
 
-    /* Store the PyGIClosure as extra args data so _pygi_marshal_cleanup_from_py_interface_callback
-     * can clean it up later for GI_SCOPE_TYPE_CALL based closures.
-     */
-    state->args_data[arg_cache->c_arg_index] = closure;
+    /* Use the PyGIClosure as data passed to cleanup for GI_SCOPE_TYPE_CALL. */
+    *cleanup_data = closure;
 
     return TRUE;
 }
@@ -1340,7 +1439,8 @@ _pygi_marshal_from_py_interface_enum (PyGIInvokeState   *state,
                                       PyGICallableCache *callable_cache,
                                       PyGIArgCache      *arg_cache,
                                       PyObject          *py_arg,
-                                      GIArgument        *arg)
+                                      GIArgument        *arg,
+                                      gpointer          *cleanup_data)
 {
     PyObject *py_long;
     long c_long;
@@ -1408,7 +1508,8 @@ _pygi_marshal_from_py_interface_flags (PyGIInvokeState   *state,
                                        PyGICallableCache *callable_cache,
                                        PyGIArgCache      *arg_cache,
                                        PyObject          *py_arg,
-                                       GIArgument        *arg)
+                                       GIArgument        *arg,
+                                       gpointer          *cleanup_data)
 {
     PyObject *py_long;
     long c_long;
@@ -1455,20 +1556,27 @@ _pygi_marshal_from_py_interface_struct_cache_adapter (PyGIInvokeState   *state,
                                                       PyGICallableCache *callable_cache,
                                                       PyGIArgCache      *arg_cache,
                                                       PyObject          *py_arg,
-                                                      GIArgument        *arg)
+                                                      GIArgument        *arg,
+                                                      gpointer          *cleanup_data)
 {
     PyGIInterfaceCache *iface_cache = (PyGIInterfaceCache *)arg_cache;
 
-    return _pygi_marshal_from_py_interface_struct (py_arg,
-                                                   arg,
-                                                   arg_cache->arg_name,
-                                                   iface_cache->interface_info,
-                                                   arg_cache->type_info,
-                                                   iface_cache->g_type,
-                                                   iface_cache->py_type,
-                                                   arg_cache->transfer,
-                                                   TRUE, /*copy_reference*/
-                                                   iface_cache->is_foreign);
+    gboolean res =  _pygi_marshal_from_py_interface_struct (py_arg,
+                                                            arg,
+                                                            arg_cache->arg_name,
+                                                            iface_cache->interface_info,
+                                                            iface_cache->g_type,
+                                                            iface_cache->py_type,
+                                                            arg_cache->transfer,
+                                                            TRUE, /*copy_reference*/
+                                                            iface_cache->is_foreign,
+                                                            arg_cache->is_pointer);
+
+    /* Assume struct marshaling is always a pointer and assign cleanup_data
+     * here rather than passing it further down the chain.
+     */
+    *cleanup_data = arg->v_pointer;
+    return res;
 }
 
 gboolean
@@ -1476,7 +1584,8 @@ _pygi_marshal_from_py_interface_boxed (PyGIInvokeState   *state,
                                        PyGICallableCache *callable_cache,
                                        PyGIArgCache      *arg_cache,
                                        PyObject          *py_arg,
-                                       GIArgument        *arg)
+                                       GIArgument        *arg,
+                                       gpointer          *cleanup_data)
 {
     PyErr_Format (PyExc_NotImplementedError,
                   "Marshalling for this type is not implemented yet");
@@ -1488,8 +1597,11 @@ _pygi_marshal_from_py_interface_object (PyGIInvokeState   *state,
                                         PyGICallableCache *callable_cache,
                                         PyGIArgCache      *arg_cache,
                                         PyObject          *py_arg,
-                                        GIArgument        *arg)
+                                        GIArgument        *arg,
+                                        gpointer          *cleanup_data)
 {
+    gboolean res = FALSE;
+
     if (py_arg == Py_None) {
         arg->v_pointer = NULL;
         return TRUE;
@@ -1509,7 +1621,9 @@ _pygi_marshal_from_py_interface_object (PyGIInvokeState   *state,
         return FALSE;
     }
 
-    return _pygi_marshal_from_py_gobject (py_arg, arg, arg_cache->transfer);
+    res = _pygi_marshal_from_py_gobject (py_arg, arg, arg_cache->transfer);
+    *cleanup_data = arg->v_pointer;
+    return res;
 }
 
 gboolean
@@ -1517,90 +1631,14 @@ _pygi_marshal_from_py_interface_union (PyGIInvokeState   *state,
                                        PyGICallableCache *callable_cache,
                                        PyGIArgCache      *arg_cache,
                                        PyObject          *py_arg,
-                                       GIArgument        *arg)
+                                       GIArgument        *arg,
+                                       gpointer          *cleanup_data)
 {
     PyErr_Format(PyExc_NotImplementedError,
                  "Marshalling for this type is not implemented yet");
     return FALSE;
 }
 
-gboolean _pygi_marshal_from_py_interface_instance (PyGIInvokeState   *state,
-                                                   PyGICallableCache *callable_cache,
-                                                   PyGIArgCache      *arg_cache,
-                                                   PyObject          *py_arg,
-                                                   GIArgument        *arg)
-{
-    GIInfoType info_type;
-    PyGIInterfaceCache *iface_cache = (PyGIInterfaceCache *)arg_cache;
-
-    info_type = g_base_info_get_type (iface_cache->interface_info);
-    switch (info_type) {
-        case GI_INFO_TYPE_UNION:
-        case GI_INFO_TYPE_STRUCT:
-        {
-            GType type = iface_cache->g_type;
-
-            if (!PyObject_IsInstance (py_arg, iface_cache->py_type)) {
-                /* wait, we might be a member of a union so manually check */
-                if (!_is_union_member (iface_cache->interface_info, py_arg)) {
-                    if (!PyErr_Occurred()) {
-                        PyObject *module = PyObject_GetAttrString(py_arg, "__module__");
-                        PyErr_Format (PyExc_TypeError,
-                                      "argument %s: Expected a %s, but got %s%s%s",
-                                      arg_cache->arg_name ? arg_cache->arg_name : "self",
-                                      iface_cache->type_name,
-                                      module ? PYGLIB_PyUnicode_AsString(module) : "",
-                                      module ? "." : "",
-                                      py_arg->ob_type->tp_name);
-                        if (module)
-                            Py_DECREF (module);
-                    }
-                    return FALSE;
-                }
-            }
-
-            if (g_type_is_a (type, G_TYPE_BOXED)) {
-                arg->v_pointer = pyg_boxed_get (py_arg, void);
-            } else if (g_type_is_a (type, G_TYPE_POINTER) ||
-                           g_type_is_a (type, G_TYPE_VARIANT) ||
-                               type == G_TYPE_NONE) {
-                arg->v_pointer = pyg_pointer_get (py_arg, void);
-            } else {
-                 PyErr_Format (PyExc_TypeError, "unable to convert an instance of '%s'", g_type_name (type));
-                 return FALSE;
-            }
-
-            break;
-        }
-        case GI_INFO_TYPE_OBJECT:
-        case GI_INFO_TYPE_INTERFACE:
-            arg->v_pointer = pygobject_get (py_arg);
-            if (arg->v_pointer != NULL) {
-                GType obj_type = G_OBJECT_TYPE (( GObject *)arg->v_pointer);
-                GType expected_type = iface_cache->g_type;
-
-                if (!g_type_is_a (obj_type, expected_type)) {
-                    PyObject *module = PyObject_GetAttrString(py_arg, "__module__");
-                    PyErr_Format (PyExc_TypeError, "argument %s: Expected %s, but got %s%s%s",
-                                  arg_cache->arg_name ? arg_cache->arg_name : "self",
-                                  iface_cache->type_name,
-                                  module ? PYGLIB_PyUnicode_AsString(module) : "",
-                                  module ? "." : "",
-                                  py_arg->ob_type->tp_name);
-                    if (module)
-                        Py_DECREF (module);
-                    return FALSE;
-                }
-            }
-            break;
-        default:
-            /* Other types don't have methods. */
-            g_assert_not_reached ();
-   }
-
-   return TRUE;
-}
-
 /* _pygi_marshal_from_py_gobject:
  * py_arg: (in):
  * arg: (out):
@@ -1773,12 +1811,12 @@ _pygi_marshal_from_py_interface_struct (PyObject *py_arg,
                                         GIArgument *arg,
                                         const gchar *arg_name,
                                         GIBaseInfo *interface_info,
-                                        GITypeInfo *type_info,
                                         GType g_type,
                                         PyObject *py_type,
                                         GITransfer transfer,
                                         gboolean copy_reference,
-                                        gboolean is_foreign)
+                                        gboolean is_foreign,
+                                        gboolean is_pointer)
 {
     gboolean is_union = FALSE;
 
@@ -1834,7 +1872,7 @@ _pygi_marshal_from_py_interface_struct (PyObject *py_arg,
     } else if (g_type_is_a (g_type, G_TYPE_POINTER) ||
                g_type_is_a (g_type, G_TYPE_VARIANT) ||
                g_type  == G_TYPE_NONE) {
-        g_warn_if_fail (g_type_is_a (g_type, G_TYPE_VARIANT) || !g_type_info_is_pointer (type_info) || transfer == GI_TRANSFER_NOTHING);
+        g_warn_if_fail (g_type_is_a (g_type, G_TYPE_VARIANT) || !is_pointer || transfer == GI_TRANSFER_NOTHING);
 
         if (g_type_is_a (g_type, G_TYPE_VARIANT) &&
                 pyg_type_from_object (py_arg) != G_TYPE_VARIANT) {
@@ -1842,6 +1880,9 @@ _pygi_marshal_from_py_interface_struct (PyObject *py_arg,
             return FALSE;
         }
         arg->v_pointer = pyg_pointer_get (py_arg, void);
+        if (transfer == GI_TRANSFER_EVERYTHING) {
+            g_variant_ref ((GVariant *)arg->v_pointer);
+        }
 
     } else {
         PyErr_Format (PyExc_NotImplementedError,
